(function() {
  // patches.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A patches utility which manages the automatic client updates.

  // This package is responsible for tracking which patches are available
  // for which versions of the product, and delivering those patches when
  // requested. In each version, one or more patches might be available:
  // for example, a different patch might be delivered to swiftpage-branded
  // clients than is delivered to ce-branded clients. When a client
  // requests an update, it sends its vitae (which includes such things as
  // its platform, arch, release, brand, and version), and this package
  // determines which single patch (if any) is best suited for that client.

  // The patch selected is the "most specific" patch: that is, the one with
  // more matching tags than any other patch. For example, if a client
  // vitae has {platform: 'win32', brand: 'swiftpage'}; and there are two
  // patches available: {platform: 'win32'} and {platform: 'win32', brand:
  // 'swiftpage'}, then the second patch would be selected, since more of
  // its tags match.

  // The idea of a "most specific" patch introduces the idea of
  // "ambiguity". For example, there may be a patch {platform: 'win32'} and
  // another patch {brand: 'swiftpage'}. If there are no other patches,
  // this creates an ambiguity: what if a client is both swiftpage-branded
  // *and* windows? Both patches match equally, so there is no "most
  // specific" patch. To resolve this ambiguity, we need to introduce a
  // third patch that has both tags {platform: 'win32', brand: 'swiftpage'}
  // which matches better than either of the other two.

  // We use set theory to detect ambiguity, describing a patch as
  // set of all possible client vitae to which the patch could apply, and
  // thus the "most specific" patch as patch which is a strict subset of
  // all other candidates.

  //     Paired Ambiguity Theorem: there exists no ambiguity if and only if
  //     for every pair of patches p1 and p2, one of the following is true:

  //     * p1 and p2 do not overlap (i.e. intersection(p1, p2) == empty())

  //     * p1 is a subset of p2 (or vice versa)

  //     * There exists some patch p3 (which is not p1 or p2) that matches
  //       the intersection of p1 and p2 exactly
  //       (i.e. intersection(p1, p2) = p3)

  // Once we can detect ambiguity, we can then detect conflicts easily:
  // There is no conflict if and only if there is no ambiguity, *and* there
  // are no two patches which correspond to exactly the same set.

  // The 'addpatch' server command should fail if it would otherwise
  // introduce any ambiguity or conflict.
  var _, async, createManager, forge, fs, init, mkdirp, patchAdjust, patchCompare, patchEquals, patchFind, patchGet, patchIntersect, patchPathsFrom, path;

  _ = require('lodash');

  async = require('async');

  fs = require('fs');

  path = require('path');

  mkdirp = require('mkdirp');

  forge = require('node-forge');

  patchPathsFrom = function(configPath) {
    var dpath, ipath;
    dpath = path.join(path.dirname(configPath), 'patches');
    ipath = path.join(dpath, 'index.json');
    return [dpath, ipath];
  };

  exports.init = init = function(configPath, configData, cb) {
    var indexPath, patchPath;
    [patchPath, indexPath] = patchPathsFrom(configPath);
    return mkdirp(patchPath).then((err) => {
      if (err != null) {
        return cb(err);
      }
      return fs.exists(indexPath, function(exists) {
        var data;
        if (exists) {
          return cb(null, false, indexPath);
        } else {
          data = {
            versions: {}
          };
          //   * '1.2.4':
          //       * file: 'randomgen3.patch'
          //         size: 182982
          //         name: '1.2.4 -> 1.2.5'
          //         tags: {}
          //       ...
          //   * '1.2.3':
          //       * file: 'randomgen2.patch'
          //         size: 182982
          //         name: '1.2.3 -> 1.2.4'
          //         tags: {}
          //       * file: 'randomgen9.patch'
          //         size: 192831
          //         name: '1.2.3 -> 1.2.4'
          //         tags:
          //           platform: 'win32'
          //       * file: 'randomgen.patch'
          //         size: 140392
          //         name: '1.2.3 -> 1.2.4'
          //         tags:
          //           platform: 'win32'
          //           brand: 'swiftpage'
          return fs.writeFile(indexPath, JSON.stringify(data, null, 2), {
            encoding: 'utf8'
          }, function(err) {
            return cb(err, true, indexPath);
          });
        }
      });
    });
  };

  exports.manage = function(configPath, configData, cb) {
    return init(configPath, configData, function(err) {
      var indexPath, patchPath;
      if (err != null) {
        return cb(err);
      }
      [patchPath, indexPath] = patchPathsFrom(configPath);
      return fs.readFile(indexPath, {
        encoding: 'utf8'
      }, function(err, data) {
        if (err != null) {
          return cb(err);
        }
        data = JSON.parse(data);
        return cb(null, createManager(indexPath, data));
      });
    });
  };

  createManager = function(ipath, data) {
    var check, readable, retrieve, verify, verifyAll, write;
    write = function(cb) {
      return fs.writeFile(ipath, JSON.stringify(data, null, 2), {
        encoding: 'utf8'
      }, cb);
    };
    verifyAll = function(cb) {
      return async.each(_.toPairs(data.versions), (function([k, v], cb) {
        return verify(k, _.cloneDeep(v), cb);
      }), cb);
    };
    //# Verify there are no conflicts or ambiguities
    //# The second argument, `vdata` will be re-sorted, so you may want to
    //# pass in a clone.
    verify = function(version, vdata, cb) {
      var ambigs, confls, found, i, isect, j, k, l, m, n, ref, ref1, ref2, ref3, ref4;
      // to check for ambiguity: sort patches topographically, so that
      // (p1 subset p2 => p1 > p2). Then check pairs, searching for
      // p3 > p1, p3 > p2
      vdata.sort(patchCompare);
      ambigs = [];
      confls = [];
      for (i = l = 0, ref = vdata.length; (0 <= ref ? l < ref : l > ref); i = 0 <= ref ? ++l : --l) {
        for (j = m = ref1 = i + 1, ref2 = vdata.length; (ref1 <= ref2 ? m < ref2 : m > ref2); j = ref1 <= ref2 ? ++m : --m) {
          if (patchEquals(vdata[i], vdata[j])) {
            if (!(_.some(confls, function(a) {
              return patchEquals({
                tags: a
              }, vdata[i]);
            }))) {
              confls.push(vdata[i].tags);
            }
            continue;
          }
          if (patchCompare(vdata[i], vdata[j]) < 0) {
            continue;
          }
          isect = patchIntersect(vdata[i], vdata[j]);
          if (isect.empty) {
            continue;
          }
          found = false;
          for (k = n = ref3 = j + 1, ref4 = vdata.length; (ref3 <= ref4 ? n < ref4 : n > ref4); k = ref3 <= ref4 ? ++n : --n) {
            if (patchEquals(isect, vdata[k])) {
              found = true;
              break;
            }
          }
          if (found) {
            continue;
          }
          if (!(_.some(ambigs, function(a) {
            return patchEquals({
              tags: a
            }, isect);
          }))) {
            ambigs.push(isect.tags);
          }
        }
      }
      if (ambigs.length === 0 && confls.length === 0) {
        return cb(null);
      } else if (ambigs.length === 0) {
        return cb(new Error(`conflicts in ${version}: ${JSON.stringify(confls)}`));
      } else if (confls.length === 0) {
        return cb(new Error(`ambiguities in ${version}: ${JSON.stringify(ambigs)}`));
      } else {
        return cb(new Error(`in ${version} ambiguities: ${JSON.stringify(ambigs)}, conflicts: ${JSON.stringify(confls)}`));
      }
    };
    check = function(vitae, cb) {
      var ref, vdata;
      vdata = (ref = data.versions[vitae.version]) != null ? ref : [];
      return patchFind(vdata, vitae, function(err, patch) {
        var result;
        if (err != null) {
          return cb(err);
        }
        if (patch != null) {
          result = _.cloneDeep(patch);
          result.version = vitae.version;
          result.update = true;
        } else {
          result = {
            update: false
          };
        }
        return cb(null, result);
      });
    };
    retrieve = function(version, id, cb) {
      return patchGet(data.versions[version], id, cb);
    };
    readable = function(patch, done) {
      var read;
      read = fs.createReadStream('/Users/charlieozinga/tmp/IN_PATCH');
      read.once('end', done);
      return read;
    };
    return {
      check: check,
      verifyAll: verifyAll,
      verify: verify,
      retrieve: retrieve,
      readable: readable
    };
  };

  // Returns -1 if p1 is a strict superset (or possible disjoint) of p2; 1
  // if p2 is a strict superset (or possible disjoint) of p1; 0 otherwise.
  patchCompare = function(p1, p2) {
    var p1only, p2only;
    p1only = _.difference(_.keys(p1.tags), _.keys(p2.tags));
    p2only = _.difference(_.keys(p2.tags), _.keys(p1.tags));
    if (_.isEmpty(p1only)) {
      if (_.isEmpty(p2only)) {
        return 0;
      } else {
        return -1;
      }
    } else if (_.isEmpty(p2only)) {
      return 1;
    } else {
      return 0;
    }
  };

  // Returns true if and only if p1 and p2 have the same values for the
  // same tags.
  patchEquals = function(p1, p2) {
    var k, l, len, p1only, p2only, ref;
    p1only = _.difference(_.keys(p1.tags), _.keys(p2.tags));
    p2only = _.difference(_.keys(p2.tags), _.keys(p1.tags));
    if (!_.isEmpty(p1only) || !_.isEmpty(p2only)) {
      return false;
    }
    ref = _.keys(p1.tags);
    for (l = 0, len = ref.length; l < len; l++) {
      k = ref[l];
      if (p1.tags[k] !== p2.tags[k]) {
        return false;
      }
    }
    return true;
  };

  // Returns an intersect object whose tags are the union of all tags from
  // both patches (i.e., is the intersection of the patch sets), or {empty:
  // true} if the patches are disjoint.
  patchIntersect = function(p1, p2) {
    var isect, k, l, len, len1, len2, m, n, p1only, p2only, pboth;
    p1only = _.difference(_.keys(p1.tags), _.keys(p2.tags));
    p2only = _.difference(_.keys(p2.tags), _.keys(p1.tags));
    pboth = _.intersection(_.keys(p1.tags), _.keys(p2.tags));
    isect = {
      empty: false,
      tags: {}
    };
    for (l = 0, len = pboth.length; l < len; l++) {
      k = pboth[l];
      if (p1.tags[k] !== p2.tags[k]) {
        return {
          empty: true
        };
      }
      isect.tags[k] = p1.tags[k];
    }
    for (m = 0, len1 = p1only.length; m < len1; m++) {
      k = p1only[m];
      isect.tags[k] = p1.tags[k];
    }
    for (n = 0, len2 = p2only.length; n < len2; n++) {
      k = p2only[n];
      isect.tags[k] = p2.tags[k];
    }
    return isect;
  };

  // Find the "most specific" patch for a given vitae.
  patchFind = function(vdata, vitae, cb) {
    var best, candidates;
    candidates = _.filter(vdata, function(p) {
      var both, k, l, len, ponly;
      ponly = _.difference(_.keys(p.tags), _.keys(vitae));
      if (!(_.isEmpty(ponly))) {
        return false;
      }
      both = _.intersection(_.keys(p.tags), _.keys(vitae));
      for (l = 0, len = both.length; l < len; l++) {
        k = both[l];
        if (p.tags[k] !== vitae[k]) {
          return false;
        }
      }
      return true;
    });
    best = _.maxBy(candidates, function(c) {
      return _.keys(c.tags).length;
    });
    return patchAdjust(best, cb);
  };

  // Find the "most specific" patch for a given vitae.
  patchGet = function(vdata, id, cb) {
    var best;
    best = _.find(vdata, function(p) {
      return p.id === id;
    });
    if (best != null) {
      return patchAdjust(best, cb);
    } else {
      return cb(new Error(`Couldn't retrieve patch ${id}`));
    }
  };

  patchAdjust = function(patch, cb) {
    if (patch == null) {
      return cb(null);
    }
    return fs.stat('/Users/charlieozinga/tmp/IN_PATCH', function(err, stats) {
      if (err != null) {
        return cb(err);
      }
      patch = _.cloneDeep(patch);
      patch.size = stats.size;
      return cb(null, patch);
    });
  };

}).call(this);
