(function() {
  // registry_es.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A registration utility to provide some registry management functions
  // for the server. Uses elasticsearch.

  // Currently, we use a single ES index named 'g2c' for all operations,
  // with the following types:

  // "usage": a time-series request log document generated by the HTTP
  // proxy with the following data:
  // {
  //   timestamp: long             # the timestamp
  //   date: string                # the display version of the timestamp
  //   available: boolean          # if the service was available
  //   success: boolean            # whether the request succeeded
  //   regId: string               # the registration ID of the client
  //   statusCode: int (optional)  # the response HTTP code
  //   reason: string (optional)   # the reason for a failure
  //   req: object                 # the request (or some portion)
  //     headers: object           # the request headers
  //   res: object                 # the response (or some portion)
  //     headers: object           # the response headers
  // }
  var _, async, createManager, elastic, initializeEs;

  _ = require('lodash');

  elastic = require('elasticsearch');

  async = require('async');

  exports.init = function(configPath, configData, cb) {
    var esUrl, ref;
    esUrl = configData != null ? (ref = configData.server) != null ? ref.esUrl : void 0 : void 0;
    return initializeEs(esUrl, cb);
  };

  initializeEs = function(esUrl, cb) {
    var esClient;
    if (esUrl == null) {
      esUrl = 'elastic://localhost:9200';
    }
    if (!esUrl.startsWith('elastic://')) {
      return cb(`Unknown stats type ${esUrl}`);
    }
    esUrl = esUrl.substr("elastic://".length);
    esClient = new elastic.Client({
      host: esUrl,
      log: 'warn'
    });
    return cb(null);
  };

  exports.manage = function(configPath, configData, cb) {
    var esUrl, ref;
    esUrl = configData != null ? (ref = configData.server) != null ? ref.esUrl : void 0 : void 0;
    return createManager(esUrl, cb);
  };

  createManager = function(esUrl, cb) {
    var counts, esClient, usage, used;
    if (esUrl == null) {
      esUrl = 'elastic://localhost:9200';
    }
    if (!esUrl.startsWith('elastic://')) {
      return cb(`Unknown stats type ${esUrl}`);
    }
    esUrl = esUrl.substr("elastic://".length);
    esClient = new elastic.Client({
      host: esUrl,
      log: 'warn'
    });
    used = function(id, brand, score, data, cb) {
      data.timestamp = score;
      data.date = new Date(score).toString();
      data.regId = id;
      data.brand = brand;
      return esClient.index({
        index: 'g2c',
        type: 'usage',
        body: data
      }, cb);
    };
    usage = function(id, start, end, cb) {
      return esClient.search({
        index: 'g2c',
        type: 'usage',
        body: {
          query: {
            range: {
              timestamp: {
                gte: start,
                lte: end
              }
            },
            term: {
              regId: id
            }
          }
        }
      }, cb);
    };
    counts = function(type, code, brand, start, end, cb) {
      var search, total;
      total = 0;
      search = ['successes'];
      if (type === 'unavailable') {
        search = ['unavailables'];
      } else if (type === 'error') {
        search = ['errors'];
      } else if (type === 'success') {
        search = ['successes'];
      } else if (type === 'errors') {
        search = ['unavailables', 'errors'];
      } else if (type === 'requests') {
        search = ['errors', 'unavailables', 'successes'];
      }
      return async.forEachSeries(search, function(type, cb) {
        if (code != null) {
          return esClient.zrangebyscore(`g2c/${type}/${brand}`, start, end, function(err, dats) {
            var filtered;
            if (err != null) {
              return cb(err);
            }
            filtered = _.filter(dats, function(d) {
              var parsed;
              try {
                parsed = JSON.parse(d);
                return ('' + parsed.code) === code;
              } catch (error) {
                return false;
              }
            });
            total += filtered.length;
            return cb();
          });
        } else {
          return esClient.zcount(`g2c/${type}/${brand}`, start, end, function(err, dats) {
            if (err != null) {
              return cb(err);
            }
            total += dats;
            return cb();
          });
        }
      }, function(err) {
        return cb(err, total);
      });
    };
    return cb(null, {
      used: used,
      usage: usage,
      counts: counts,
      quit: function() {
        return esClient.close();
      }
    });
  };

}).call(this);
