(function() {
  // registry_redis.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A registration utility to provide registry management functions for
  // the server. Uses redis.

  // The following are the keys used in redis to track registrations:

  // "g2c/allids": a set containing all the registration IDs
  // "g2c/openPorts": the set of all ports still available
  // "g2c/max/id": the maximum registered ID

  // per registration ID:
  // "g2c/ids/{regId}/vitae": a JSON string with vitae data.
  // "g2c/ids/{regId}/aliases": a set containing all aliases.
  // "g2c/ids/{regId}/key": the public key PEM
  // "g2c/ids/{regId}/usage": a sorted set (score = timestamp) of usage
  // "g2c/ids/{regId}/port": the port number for this ID
  // "g2c/ports/{port}": the registration ID (for reverse lookup)
  // "g2c/aliases/{alias}": the registration ID (for reverse lookup)

  // per brand
  // "g2c/registers/{brand}": a sorted set (score = timestamp) of registrations
  // "g2c/errors/{brand}": a sorted set (score = timestamp) of error responses
  // "g2c/unavailables/{brand}": a sorted set of unavailable responses
  // "g2c/successes/{brand}": a sorted set of successful responses
  var _, akPath, async, createAuthkeys, createManager, forge, fs, home, mkdirp, path, redis, shPath;

  _ = require('lodash');

  redis = require('redis');

  async = require('async');

  fs = require('fs');

  path = require('path');

  mkdirp = require('mkdirp');

  forge = require('node-forge');

  home = process.env.HOME || process.env.USERPROFILE;

  shPath = path.join(home, '.ssh');

  akPath = path.join(shPath, 'authorized_keys');

  exports.init = function(configPath, configData, cb) {
    var cb2, ref, regUrl;
    regUrl = configData != null ? (ref = configData.server) != null ? ref.registryfile : void 0 : void 0;
    if (regUrl == null) {
      return cb(new Error("No registryUrl in configuration"));
    }
    cb2 = createAuthkeys(regUrl, cb);
    return cb2(null, true);
  };

  createAuthkeys = function(regUrl, cb) {
    return function(err, made) {
      if (err != null) {
        return cb(err, made, regUrl);
      }
      return fs.exists(akPath, function(exists) {
        if (exists) {
          return cb(null, made, regUrl);
        }
        return mkdirp(shPath, 0o700, function(err) {
          if (err != null) {
            return cb(err, made, regUrl);
          }
          return fs.writeFile(akPath, '', {
            encoding: 'ascii',
            mode: 0o600
          }, function(err) {
            return cb(err, made, regUrl);
          });
        });
      });
    };
  };

  exports.manage = function(configPath, configData, cb) {
    var ref, regUrl;
    regUrl = configData != null ? (ref = configData.server) != null ? ref.registryUrl : void 0 : void 0;
    if (regUrl == null) {
      return cb(new Error("No registryUrl in configuration"));
    }
    return cb(null, createManager(regUrl));
  };

  createManager = function(redisUrl) {
    var addAlias, addAuthorized, connected, connecting, connects, counts, disconnected, each, find, findAlias, findFromPref, findPortFromPref, get, isAvailable, isConnected, pickPort, rclient, register, registers, removeAlias, removeAuthorized, sameKey, setVitae, setnxGoodId, setnxGoodPort, size, status, statusOf, unregister, usage, used;
    if (redisUrl == null) {
      redisUrl = 'redis://localhost:6379';
    }
    connects = {};
    rclient = redis.createClient(redisUrl);
    status = function(cb) {
      return rclient.smembers('g2c/allids', function(err, mems) {
        var regId;
        if (err != null) {
          return cb(err);
        }
        return cb(null, {
          status: 'running',
          connections: (function() {
            var i, len, ref, results;
            results = [];
            for (i = 0, len = mems.length; i < len; i++) {
              regId = mems[i];
              results.push({
                registeredId: regId,
                connected: (ref = connects[regId]) != null ? ref : false
              });
            }
            return results;
          })()
        });
      });
    };
    statusOf = function(regId, cb) {
      var keys;
      keys = ['vitae', 'key', 'port'].map(function(p) {
        return `g2c/ids/${regId}/${p}`;
      });
      return rclient.mget(keys, function(err, vals) {
        var fpt, pubkey, vitae;
        if (err != null) {
          return cb(err);
        }
        pubkey = forge.pki.publicKeyFromPem(vals[1]);
        fpt = forge.ssh.getPublicKeyFingerprint(pubkey, {
          encoding: 'hex',
          delimiter: ':'
        });
        vitae = (function() {
          try {
            return JSON.parse(vals[0]);
          } catch (error) {
            return {};
          }
        })();
        return rclient.smembers(`g2c/ids/${regId}/aliases`, function(err, als) {
          var ref;
          if (err != null) {
            return cb(err);
          }
          return cb(null, {
            status: 'running',
            connections: [
              {
                registeredId: regId,
                port: parseInt(vals[2]),
                fingerprint: fpt,
                connected: (ref = connects[regId]) != null ? ref : false,
                vitae: vitae,
                aliases: als
              }
            ]
          });
        });
      });
    };
    addAlias = function(regId, alias, cb) {
      return rclient.exists(`g2c/aliases/${alias}`, function(err, ex) {
        if (ex) {
          return cb(`Alias ${alias} already exists.`);
        }
        return rclient.sismember('g2c/allids', regId, function(err, ex) {
          if (!ex) {
            return cb(`No such registeredId ${regId}.`);
          }
          return rclient.sadd("g2c/ids/{regId}/aliases", alias, function(err) {
            if (err != null) {
              return cb(err);
            }
            return rclient.set(`g2c/aliases/${alias}`, regId, cb);
          });
        });
      });
    };
    removeAlias = function(regId, alias, cb) {
      return rclient.sismember(`g2c/ids/${regId}/aliases`, alias, function(err, ism) {
        if (!ism) {
          return cb(`${alias} is not an alias of ${regId}`);
        }
        return rclient.srem(`g2c/ids/${regId}/aliases`, alias, function(err) {
          if (err != null) {
            return cb(err);
          }
          return rclient.del(`g2c/aliases/${alias}`, cb);
        });
      });
    };
    sameKey = function(regIds, cb) {
      var commonfpt, keys;
      commonfpt = null;
      keys = regIds.map(function(r) {
        return `g2c/ids/${r}/key`;
      });
      return rclient.mget(keys, function(err, vals) {
        var e, fpt, i, len, pubkey, val;
        if (err != null) {
          return cb(err);
        }
        for (i = 0, len = vals.length; i < len; i++) {
          val = vals[i];
          try {
            pubkey = forge.pki.publicKeyFromPem(val);
            fpt = forge.ssh.getPublicKeyFingerprint(pubkey, {
              encoding: 'hex',
              delimiter: ':'
            });
          } catch (error) {
            e = error;
            return cb(e);
          }
          if (commonfpt == null) {
            commonfpt = fpt;
          } else if (commonfpt !== fpt) {
            return cb(null, false);
          }
        }
        return cb(null, true);
      });
    };
    setVitae = function(regId, vitae, cb) {
      console.log("setting vitae", vitae);
      return rclient.sismember('g2c/allids', regId, function(err, ex) {
        if (!ex) {
          return typeof cb === "function" ? cb(`No such registeredId ${regId}.`) : void 0;
        }
        if (vitae != null) {
          return rclient.set(`g2c/ids/${regId}/vitae`, JSON.stringify(vitae), function(err) {
            var brand, jdata, ref;
            if (err != null) {
              return typeof cb === "function" ? cb(err) : void 0;
            }
            brand = (ref = vitae.brand) != null ? ref : 'ce';
            jdata = JSON.stringify({
              brand: brand,
              regId: regId
            });
            return rclient.zrank(`g2c/registers/${brand}`, jdata, function(err, rnk) {
              var score;
              if (err != null) {
                return typeof cb === "function" ? cb(err) : void 0;
              }
              if (rnk == null) {
                score = new Date().getTime();
                return rclient.zadd(`g2c/registers/${brand}`, score, jdata, cb);
              }
            });
          });
        } else {
          return rclient.del(`g2c/ids/${regId}/vitae`, cb);
        }
      });
    };
    findAlias = function(alias, cb) {
      return rclient.get(`g2c/aliases/${alias}`, cb);
    };
    isAvailable = function(regIds) {
      var i, len, ref, regId, takens;
      takens = [];
      for (i = 0, len = regIds.length; i < len; i++) {
        regId = regIds[i];
        if (!(((ref = connects[regId]) != null ? ref : false) === false)) {
          takens.push(regId);
        }
      }
      return takens;
    };
    isConnected = function(regId) {
      var ref;
      return ((ref = connects[regId]) != null ? ref : false) === true;
    };
    connected = function(regId) {
      return connects[regId] = true;
    };
    connecting = function(regIds) {
      var i, len, regId, results;
      results = [];
      for (i = 0, len = regIds.length; i < len; i++) {
        regId = regIds[i];
        results.push(connects[regId] = "connecting");
      }
      return results;
    };
    disconnected = function(regId) {
      return connects[regId] = false;
    };
    size = function(cb) {
      return rclient.scard('g2c/allids', cb);
    };
    get = function(regId, cb) {
      var keys;
      keys = ['vitae', 'key', 'port'].map(function(p) {
        return `g2c/ids/${regId}/${p}`;
      });
      return rclient.mget(keys, function(err, vals) {
        var entry;
        if (err != null) {
          return cb(err);
        }
        if ((vals[2] == null) || (vals[1] == null)) {
          return cb(`No such registered ID \"${regId}\"`);
        }
        entry = {
          port: parseInt(vals[2]),
          key: vals[1]
        };
        try {
          entry.vitae = JSON.parse(vals[0]);
        } catch (error) {}
        return rclient.smembers(`g2c/ids/${regId}/aliases`, function(err, als) {
          if (err != null) {
            return cb(err);
          }
          if ((als != null) && als.length) {
            entry.aliases = als;
          }
          return cb(null, entry);
        });
      });
    };
    find = function(regIds, cb) {
      var pkeys;
      if ((regIds == null) || regIds.length === 0) {
        return cb(null, null, []);
      }
      pkeys = regIds.map(function(r) {
        return `g2c/ids/${r}/port`;
      });
      return rclient.mget(pkeys, function(err, pvals) {
        if (err != null) {
          return cb(err);
        }
        if (_.some(pvals, function(v) {
          return v == null;
        })) {
          return cb(`Ports missing for registeredIds ${regIds}`);
        }
        return rclient.get(`g2c/ids/${regIds[0]}/key`, function(err, kval) {
          if (err != null) {
            return cb(err);
          }
          if (kval == null) {
            return cb(`Key missing for ${regIds[0]}`);
          }
          return cb(null, kval, pvals);
        });
      });
    };
    each = function(icb, cb) {
      if (cb == null) {
        cb = function() {};
      }
      return rclient.smembers('g2c/allids', function(err, ids) {
        var i, len, regId;
        if (err != null) {
          return icb(err);
        }
        for (i = 0, len = ids.length; i < len; i++) {
          regId = ids[i];
          statusOf(regId, function(err, stat) {
            var conn;
            if (err != null) {
              return icb(err);
            }
            conn = stat.connections[0];
            conn.id = conn.registeredId;
            return icb(null, conn);
          });
        }
        return cb(null);
      });
    };
    used = function(id, brand, score, data, cb) {
      return typeof cb === "function" ? cb(null) : void 0; // Will be done by elastic
    };
    usage = function(id, start, end, cb) {
      return rclient.zrangebyscore(`g2c/ids/${id}/usage`, start, end, function(err, dats) {
        if (err != null) {
          return cb(err);
        }
        return cb(null, _.map(dats, JSON.parse));
      });
    };
    counts = function(type, code, brand, start, end, cb) {
      var search, total;
      total = 0;
      search = ['successes'];
      if (type === 'unavailable') {
        search = ['unavailables'];
      } else if (type === 'error') {
        search = ['errors'];
      } else if (type === 'success') {
        search = ['successes'];
      } else if (type === 'errors') {
        search = ['unavailables', 'errors'];
      } else if (type === 'requests') {
        search = ['errors', 'unavailables', 'successes'];
      }
      return async.forEachSeries(search, function(type, cb) {
        if (code != null) {
          return rclient.zrangebyscore(`g2c/${type}/${brand}`, start, end, function(err, dats) {
            var filtered;
            if (err != null) {
              return cb(err);
            }
            filtered = _.filter(dats, function(d) {
              var parsed;
              try {
                parsed = JSON.parse(d);
                return ('' + parsed.code) === code;
              } catch (error) {
                return false;
              }
            });
            total += filtered.length;
            return cb();
          });
        } else {
          return rclient.zcount(`g2c/${type}/${brand}`, start, end, function(err, dats) {
            if (err != null) {
              return cb(err);
            }
            total += dats;
            return cb();
          });
        }
      }, function(err) {
        return cb(err, total);
      });
    };
    registers = function(brand, start, end, cb) {
      return rclient.zcount(`g2c/registers/${brand}`, start, end, cb);
    };
    setnxGoodPort = function(port, prefId, cb) {
      if (!_.isInteger(port) && /^\d+$/.test(port)) {
        port = parseInt(port);
      }
      if (port == null) {
        return cb(null, false);
      }
      if (!_.isInteger(port)) {
        return cb(null, false);
      }
      if (port < 11081 || port > 64000) {
        return cb(null, false);
      }
      return rclient.setnx(`g2c/ports/${port}`, prefId, function(err, prv) {
        if (err != null) {
          return cb(err);
        }
        if (prv) {
          return rclient.zrem("g2c/openPorts", port, function(err) {
            return cb(err, prv);
          });
        } else {
          return cb(err, prv);
        }
      });
    };
    pickPort = function(tries, prefId, cb) {
      if (tries > 100) {
        return cb("No available port range");
      }
      return rclient.zrange('g2c/openPorts', 0, 100, function(err, rng) {
        var attempt, attemptStr, e, index;
        if (err != null) {
          return cb(err);
        }
        if (cb.length === 0) {
          return cb("No available port");
        }
        index = Math.floor(Math.random() * rng.length);
        attemptStr = rng[index];
        try {
          attempt = parseInt(attemptStr);
        } catch (error) {
          e = error;
          return cb(e);
        }
        return rclient.setnx(`g2c/ports/${attempt}`, prefId, function(err, got) {
          if (got) {
            return rclient.zrem("g2c/openPorts", attempt, function(err) {
              return cb(err, attempt);
            });
          } else {
            return pickPort(tries + 1, prefId, cb);
          }
        });
      });
    };
    findPortFromPref = function(pref, cb) {
      var obtained;
      obtained = {};
      return setnxGoodPort(pref.port, pref.id, function(err, prv) {
        if (err != null) {
          return cb(err);
        }
        if (prv) {
          obtained.port = pref.port;
          return rclient.zrem("g2c/openPorts", pref.port, function(err) {
            return cb(err, obtained);
          });
        } else {
          return pickPort(0, pref.id, function(err, maxport) {
            if (err != null) {
              return cb(err);
            }
            pref.port = obtained.port = maxport;
            return rclient.set(`g2c/ids/${pref.id}/port`, maxport, function(err) {
              return cb(err, obtained);
            });
          });
        }
      });
    };
    setnxGoodId = function(id, prefPort, cb) {
      if (prefPort == null) {
        prefPort = 0;
      }
      return rclient.setnx("g2c/max/id", 4000, function(err) {
        if (err != null) {
          return cb(err);
        }
        if (id == null) {
          return cb(null, false);
        }
        if (typeof id !== 'string') {
          return cb(null, false);
        }
        if (id.length < 4) {
          return cb(null, false);
        }
        if (/^\d+$/.test(id)) {
          return rclient.get("g2c/max/id", function(err, max) {
            if (parseInt(id) > parseInt(max)) {
              return cb(null, false);
            }
            // dangerous not to do this:
            // rclient.setnx "g2c/ids/#{id}/port", prefPort, cb
            return rclient.get(`g2c/ids/${id}/port`, function(err, oldPort) {
              if (err != null) {
                return cb(err);
              }
              if (oldPort != null) {
                return rclient.del(`g2c/ports/${oldPort}`, function(err) {
                  var now;
                  if (err != null) {
                    return cb(err);
                  }
                  now = new Date().getTime();
                  return rclient.zadd("g2c/openPorts", now, oldPort, function(err) {
                    if (err != null) {
                      return cb(err);
                    }
                    return rclient.set(`g2c/ids/${id}/port`, prefPort, cb);
                  });
                });
              } else {
                return rclient.set(`g2c/ids/${id}/port`, prefPort, cb);
              }
            });
          });
        } else {
          return rclient.get(`g2c/ids/${id}/port`, function(err, oldPort) {
            if (err != null) {
              return cb(err);
            }
            if (oldPort != null) {
              return rclient.del(`g2c/ports/${oldPort}`, function(err) {
                var now;
                if (err != null) {
                  return cb(err);
                }
                now = new Date().getTime();
                return rclient.zadd("g2c/openPorts", now, oldPort, function(err) {
                  if (err != null) {
                    return cb(err);
                  }
                  return rclient.set(`g2c/ids/${id}/port`, prefPort, cb);
                });
              });
            } else {
              return rclient.set(`g2c/ids/${id}/port`, prefPort, cb);
            }
          });
        }
      });
    };
    findFromPref = function(pref, cb) {
      var obtained;
      obtained = {};
      return setnxGoodId(pref.id, pref.port, function(err, irv) {
        if (err != null) {
          return cb(err);
        }
        if (irv) {
          obtained.id = pref.id;
          return findPortFromPref(pref, function(err, pobt) {
            if (err != null) {
              return cb(err);
            }
            obtained.port = pobt.port;
            return rclient.set(`g2c/ids/${obtained.id}/port`, obtained.port, function(err) {
              return cb(err, obtained);
            });
          });
        } else {
          return rclient.incr("g2c/max/id", function(err, maxid) {
            var ref;
            if (err != null) {
              return cb(err);
            }
            pref.id = obtained.id = `${maxid}`;
            return rclient.set(`g2c/ids/${maxid}/port`, (ref = pref.port) != null ? ref : 0, function(err) {
              if (err != null) {
                return cb(err);
              }
              return findPortFromPref(pref, function(err, pobt) {
                if (err != null) {
                  return cb(err);
                }
                obtained.port = pobt.port;
                return cb(null, obtained);
              });
            });
          });
        }
      });
    };
    register = function(pubkeypem, preferred, cb) {
      var prefclone;
      prefclone = _.cloneDeep(preferred);
      return findFromPref(preferred != null ? preferred : {}, function(err, obtained) {
        var pubkey, sshkey;
        if (err != null) {
          return cb(err);
        }
        console.log('from pref', prefclone, obtained);
        try {
          if (pubkeypem != null) {
            pubkey = forge.pki.publicKeyFromPem(pubkeypem);
          }
          if (pubkey != null) {
            sshkey = forge.ssh.publicKeyToOpenSSH(pubkey);
          }
        } catch (error) {
          err = error;
          return cb(err);
        }
        if (sshkey == null) {
          return cb(new Error('Failed public key.'));
        }
        return rclient.set(`g2c/ids/${obtained.id}/key`, pubkeypem, function(err) {
          if (err != null) {
            return cb(err);
          }
          return rclient.sadd('g2c/allids', obtained.id, function(err) {
            var entry;
            entry = {
              port: obtained.port,
              key: pubkeypem
            };
            return addAuthorized(sshkey, obtained.port, function(err) {
              return cb(err, obtained.id, entry);
            });
          });
        });
      });
    };
    unregister = function(regIds, cb) {
      // This could probably be more parallelized if needed.
      return async.eachSeries(regIds, function(regId, cb) {
        var keys;
        keys = ['vitae', 'key', 'port', 'aliases'].map(function(p) {
          return `g2c/ids/${regId}/${p}`;
        });
        return rclient.sismember('g2c/allids', regId, function(err, ism) {
          if (!ism) {
            return cb(`No entry found for ${regId}`);
          }
          return rclient.get(`g2c/ids/${regId}/port`, function(err, port) {
            if (err != null) {
              return cb(err);
            }
            return rclient.smembers(`g2c/ids/${regId}/aliases`, function(err, als) {
              var akeys;
              akeys = als.map(function(a) {
                return `g2c/aliases/${a}`;
              });
              if (err != null) {
                return cb(err);
              }
              return rclient.del(...keys, function(err) {
                if (err != null) {
                  return cb(err);
                }
                return rclient.del(`g2c/ports/${port}`, function(err) {
                  var now;
                  if (err != null) {
                    return cb(err);
                  }
                  now = new Date().getTime();
                  return rclient.zadd("g2c/openPorts", now, port, function(err) {
                    if (err != null) {
                      return cb(err);
                    }
                    if (akeys.length) {
                      return rclient.del(...akeys, function(err) {
                        if (err != null) {
                          return cb(err);
                        }
                        return rclient.srem('g2c/allids', regId, cb);
                      });
                    } else {
                      return rclient.srem('g2c/allids', regId, cb);
                    }
                  });
                });
              });
            });
          });
        });
      }, cb);
    };
    addAuthorized = function(sshkey, port, cb) {
      return fs.appendFile(akPath, `command=\"echo 'This login is restricted.'\",no-agent-forwarding,no-X11-forwarding,permitopen=\"localhost:${port}\" ${sshkey}\n`, {
        encoding: 'utf8'
      }, cb);
    };
    removeAuthorized = function(port, cb) {
      return fs.readFile(akPath, {
        encoding: 'utf8'
      }, function(err, data) {
        var lines, origLines;
        if (err != null) {
          return cb(err);
        }
        origLines = data.split('\n');
        lines = _.reject(origLines, function(l) {
          return l.indexOf(`localhost:${port}`) > -1;
        });
        return fs.writeFile(akPath, lines.join('\n'), {
          encoding: 'utf8'
        }, cb);
      });
    };
    return {
      setVitae: setVitae,
      findAlias: findAlias,
      addAlias: addAlias,
      removeAlias: removeAlias,
      status: status,
      statusOf: statusOf,
      connected: connected,
      connecting: connecting,
      isAvailable: isAvailable,
      isConnected: isConnected,
      disconnected: disconnected,
      size: size,
      each: each,
      find: find,
      get: get,
      register: register,
      unregister: unregister,
      sameKey: sameKey,
      used: used,
      usage: usage,
      counts: counts,
      registers: registers,
      quit: function() {
        return rclient.quit();
      }
    };
  };

}).call(this);
