(function() {
  // waitfor.coffee    Copyright (c) 2016 Cloud Elements
  // All rights reserved. CLOUD ELEMENTS PROPRIETARY/CONFIDENTIAL.
  // Use is subject to license terms.

  // A two-path callback synchronizer, created like this:

  // w = waitfor.create(callback)

  // These syncronizers have three callback methods, which are intended to
  // be used by certain long-lived resources which generate 'open',
  // 'close', and 'error' (or equivalent) events, as such:

  // resource.on 'open', w.open
  // resource.on 'close', w.close
  // resource.on 'error', w.error

  // It is expected that such resources either (1) produce one or more
  // errors before they successfully open (in which case it is unlikely
  // that 'open' is ever called), or (2) emit 'open', and then later emit
  // 'close' sometime afterwards.

  // This synchronizer calls the `callback` function provided at its
  // construction either at the first error (in case 1), or when the
  // 'close' event is emitted (in case 2). Either way, once the callback is
  // triggered, the synchronizer is considered complete, and all event
  // handlers become inert.

  // The callback is called with a single boolean arg, true in case 1, and
  // false in case 2, followed by all the arguments of the events
  // themselves.

  // This synchronizer can be "branched" with `w.branch()`, which creates a
  // child synchronizer. Branches can only be created in case 2, and only
  // after 'open', but before 'close' has been emitted. Any attempt to
  // create a branch at another time will result in an inert synchronizer,
  // which is not counted as a child.

  // If live branch(es) are created, then the parent synchronizer will wait
  // until both itself, and all its children are completed (via either case
  // 1 or 2) before invoking its callback.
  var create, createFake;

  exports.create = create = function(cb) {
    var children, complete, completed, opened, savedArgs, savedErr;
    opened = false;
    completed = false;
    children = 0;
    savedErr = false;
    savedArgs = [];
    complete = function(isErr, ...args) {
      completed = true;
      savedErr = isErr;
      savedArgs = args;
      if (children === 0) {
        return cb(isErr, ...args);
      }
    };
    return {
      error: function(...args) {
        if (completed || opened) {
          return;
        }
        return complete(true, ...args);
      },
      open: function() {
        if (completed) {
          return;
        }
        return opened = true;
      },
      close: function(...args) {
        if (completed) {
          return;
        }
        return complete(false, ...args);
      },
      branch: function() {
        var child;
        if (completed || !opened) {
          return createFake();
        }
        ++children;
        child = create(function(isErr) {
          --children;
          savedErr = savedErr || isErr;
          if (completed && children === 0) {
            return cb(savedErr, ...savedArgs);
          }
        });
        return child;
      }
    };
  };

  createFake = function() {
    return {
      error: function() {},
      open: function() {},
      close: function() {},
      branch: function() {
        return createFake();
      }
    };
  };

}).call(this);
